/*
 LiPo Battery Storage Monitor - Server
 Version: 1.0
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 19/11/2017
 Last Modified: 3/12/2017
 */

#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
#include "esp8266-sendemail/sendemail.h"

#include <SPI.h>
#include "mirf.h"
#include "nRF24L01.h"

const char *ssid = "";
const char *password = "";
IPAddress ip(192,168,1,3);  //Node static IP
IPAddress gateway(192,168,1,254);
IPAddress subnet(255,255,255,0);
ESP8266WebServer server ( 80 );

uint32_t currentUptime = 0;
String tempCalcTimeStr = "";
const int led = 2;
uint8_t data_in_out[14];
uint8_t linkByte = 49; // Change this if you like

// Sensors structure
#define MAX_SENSORS 30
struct Sensor {
  uint8_t state;
  uint32_t lastseen;
};
Sensor sensorList[MAX_SENSORS];

// Voltage structure
double sensorVoltage[MAX_SENSORS][6];


void handleRoot() {
  digitalWrite ( led, 1 );
  
  String temp = "<html>\
<head>\
<title>LiPo Monitor</title>\
</head>\
<link rel='stylesheet' type='text/css' href='style.css'>\
<body>\
<div class='text-css'><p align='center'><b>Batteries</b></p></div>\
<div align='center'>\
  <table border='1' width='75%' cellspacing='0' cellpadding='0' bordercolorlight='#000000'>\
    <tr>\
      <td align='center' width='7%'><font class='normaltext-css'><b>No</b></font></td>\
      <td align='center'><font class='normaltext-css'><b>Cells</b></font></td>\
      <td align='center' width='12%'><font class='normaltext-css'><b>Status</b></font></td>\
      <td align='center' width='50%'><font class='normaltext-css'><b>Voltage</b></font></td>\
      <td align='center' width='21%'><font class='normaltext-css'><b>Last check in</b></font></td>\
    </tr>";

  for (uint8_t x = 0; x < MAX_SENSORS; x++) {
    if (sensorList[x].lastseen > 0) {

      uint8_t cells = 0;
      uint8_t cellsOk = 1;

      // Check for cell count
      for (uint8_t v = 0; v < 6; v++) {
        if (sensorVoltage[x][v] >= 1) {
          cells = v+1;
        }
      }

      // Check voltage on cells (can't test if highest cell is 0V)
      for (uint8_t v = 0; v < cells; v++) {
        if (sensorVoltage[x][v] <= 3.7 || sensorVoltage[x][v] >= 4.35) {
           cellsOk = 0;
        }
      }

      calculate_last_time(sensorList[x].lastseen);

      temp += "<tr>\
      <td align='center' width='13%' height='35'><font class='normaltext-css'>";
      temp += x;
      
      temp += "</font></td><td align='center' height='35'><font class='normaltext-css'>";
      temp += cells;
      
      temp += "</font></td><td align='center' height='35'>\
      <font class='normaltext-css'>";
      
      if (cellsOk == 0) {
        temp += "Warning";
      }
      else {
        temp += "OK";
      }
      temp += "</font></td>\
      <td align='center' height='35'><font class='normaltext-css'>";

      for (uint8_t v = 0; v < cells; v++) {
        temp += sensorVoltage[x][v];
        temp += "V";
        
        if (v != (cells-1)) {
          temp += ", ";
        }
      }

      temp += "</font></td>\
      <td align='center' height='35'><font class='normaltext-css'>";
      temp += tempCalcTimeStr;
      temp += "</font></td></tr>";
    }
  }

  temp += "</table></div><br/><br/><form method='POST' action=''><input type='submit' name='submit' value='Clear Sensors List' /></form>";
  temp += "</div></body></html>";
  
  server.send ( 200, "text/html", temp );
  digitalWrite ( led, 0 );
}

void handleCss() {
  digitalWrite ( led, 1 );
  
  String temp(F(".arm-css {\
  display: inline-block;\
  -webkit-box-sizing: content-box;\
  -moz-box-sizing: content-box;\
  box-sizing: content-box;\
  width: 16%;\
  height: 5%;\
  cursor: pointer;\
  padding: 10px 20px;\
  border: 1px solid #000000;\
  -webkit-border-radius: 3px;\
  border-radius: 3px;\
  font: normal 30px/normal 'Trebuchet MS', Helvetica, sans-serif;\
  color: rgba(255,255,255,0.9);\
  -o-text-overflow: clip;\
  text-overflow: clip;\
  background: rgba(232,126,39,0.66);\
  -webkit-box-shadow: 2px 2px 2px 0 rgba(0,0,0,0.2) ;\
  box-shadow: 2px 2px 2px 0 rgba(0,0,0,0.2) ;\
  -webkit-transition: all 300ms cubic-bezier(0.42, 0, 0.58, 1);\
  -moz-transition: all 300ms cubic-bezier(0.42, 0, 0.58, 1);\
  -o-transition: all 300ms cubic-bezier(0.42, 0, 0.58, 1);\
  transition: all 300ms cubic-bezier(0.42, 0, 0.58, 1);\
}\
.text-css {\
  -webkit-box-sizing: content-box;\
  -moz-box-sizing: content-box;\
  box-sizing: content-box;\
  border: none;\
  font: normal 40px/1 'Trebuchet MS', Helvetica, sans-serif;\
  color: black;\
  -o-text-overflow: ellipsis;\
  text-overflow: ellipsis;\
}\
.normaltext-css {\
  -webkit-box-sizing: content-box;\
  -moz-box-sizing: content-box;\
  box-sizing: content-box;\
  border: none;\
  font: normal 30px/1 'Trebuchet MS', Helvetica, sans-serif;\
  color: black;\
  -o-text-overflow: ellipsis;\
  text-overflow: ellipsis;\
}"));

  server.send ( 200, "text/html", temp );
  digitalWrite ( led, 0 );
}

void handleAdmin() {
  digitalWrite ( led, 1 );

  String temp = "<html>\
<head>\
<title>LiPo Monitor - Admin</title>\
</head>\
<link rel='stylesheet' type='text/css' href='style.css'>\
<body>";
   
  temp += "<form method='POST' action=''>\
  <p align='center'><b>Sensor Linking<br>\
 </b>Hold the link button on the sensor and then connect the power.<br>\
  Enter the sensor number you wish to assign it below and press Submit.</p>\
  <p align='center'>Sensor Number <input type='text' name='sensorno' size='4'></p>\
  <p align='center'>Check in delay (1 to 255 hours, 20% -/+) <input type='text' name='hours' size='4'></p>\
  <p align='center'><input type='submit' value='Submit' name='submit'></p>\
  </form><br/>";
  
  if (server.hasArg("sensorno")) {
    uint8_t sensorNo = server.arg("sensorno").toInt();
    uint8_t hoursDelay = server.arg("hours").toInt();
    
    data_in_out[SENSOR_NUMBER] = sensorNo;
    data_in_out[LINK_BYTE] = linkByte;
    data_in_out[HOURS_DELAY] = hoursDelay;

    // Send it a few times to make sure it's received
    for (uint8_t x = 0; x < 10; x++) {
      // Transmit the sensor number and link byte to the sensor
      NRF_TX_POWERUP; // Power up to transmitter mode
      delay(3);
      mirf_flush_rx_tx(); // Flush RX and TX FIFO
      
      mirf_CSN_lo;  // Pull down chip select
      SPI.transfer(W_TX_PAYLOAD); // Write cmd to write payload
      spi_write_data(data_in_out, mirf_PAYLOAD); // Write payload
      mirf_CSN_hi; // Pull up chip select
      
      mirf_CE_hi; // Start transmission
      delay(1);
      mirf_CE_lo;
      delay(1); // Wait for transmission
      
      while (!(mirf_status() & (1<<TX_DS))) { // Wait until packet is sent
        if (mirf_status() & (1<<MAX_RT)) { // If maximum retries reached
          mirf_config_register(STATUS, 1<<MAX_RT); // Reset maximum retries to be able to transmit again next time
        }
        delay(1);
      }
      mirf_config_register(STATUS,(1<<TX_DS)); // Reset status register

      delay(50);
    }
    
    // Go back to receiver mode
    NRF_RX_POWERUP; // Power up to receiver mode
    delay(3);
    mirf_flush_rx_tx(); // Flush RX and TX FIFO
    mirf_CE_hi; // Start listening
    
    temp += "<p align='center'>Sensor number programmed: ";
    temp += sensorNo;
    temp += "<br/>Check in delay programmed: ";
    temp += hoursDelay;
    temp += "</p>";
  }
  
  server.send ( 200, "text/html", temp );
  digitalWrite ( led, 0 );
}

void handleNotFound() {
  digitalWrite ( led, 1 );
  
  String message = "File Not Found\n\n";
  message += "URI: ";
  message += server.uri();
  message += "\nMethod: ";
  message += ( server.method() == HTTP_GET ) ? "GET" : "POST";
  message += "\nArguments: ";
  message += server.args();
  message += "\n";

  for ( uint8_t i = 0; i < server.args(); i++ ) {
    message += " " + server.argName ( i ) + ": " + server.arg ( i ) + "\n";
  }

  server.send ( 404, "text/plain", message );
  digitalWrite ( led, 0 );
}


void setup ( void ) {
  pinMode ( led, OUTPUT );
  digitalWrite ( led, 0 );
  Serial.begin ( 115200 );

  WiFi.mode(WIFI_STA);
  WiFi.begin ( ssid, password );
  WiFi.config(ip, gateway, subnet);
  
  // Wait for connection
  while ( WiFi.status() != WL_CONNECTED ) {
    delay ( 500 );
  }

  server.on ( "/", handleRoot );
  server.on ( "/admin", handleAdmin );
  server.on ( "/style.css", handleCss );
  server.onNotFound ( handleNotFound );
  server.begin();
  
  mirf_setup();
  
  Serial.println("\nReady");

  noInterrupts();
  timer0_isr_init();
  timer0_attachInterrupt(timer0_ISR);
  timer0_write(ESP.getCycleCount() + 80000000L); // 80MHz = 1 sec
  interrupts();
}

// Increment uptime every second
void timer0_ISR (void) {
  timer0_write(ESP.getCycleCount() + 80000000L); // 80MHz = 1 sec
  currentUptime++;
}

void calculate_last_time(uint32_t sensorTime) {  
  uint32_t calcTime = currentUptime - sensorTime;
  uint8_t calcDays = 0;
  uint8_t calcHours = 0;
  uint8_t calcMinutes = 0;
  uint8_t calcSeconds = 0;
  
  tempCalcTimeStr = "";

  if (sensorTime >= 1) {
    if (calcTime / 86400 >= 1) {
      calcDays = calcTime / 86400;
      calcTime = calcTime - (86400 * calcDays);
      tempCalcTimeStr += calcDays;
      tempCalcTimeStr += "d, ";
    }
    if (calcTime / 3600 >= 1) {
      calcHours = calcTime / 3600;
      calcTime = calcTime - (3600 * calcHours);
      tempCalcTimeStr += calcHours;
      tempCalcTimeStr += "h, ";
    }
    if (calcTime / 60 >= 1) {
      calcMinutes = calcTime / 60;
      calcTime = calcTime - (60 * calcMinutes);
      tempCalcTimeStr += calcMinutes;
      tempCalcTimeStr += "m, ";
    }
    calcSeconds = calcTime;
    
    tempCalcTimeStr += calcSeconds;
    tempCalcTimeStr += "s";
  }
}

void loop ( void ) {
  server.handleClient();

  // Check if we have a packet waiting
  if (mirf_status() & (1<<RX_DR)) { // Wait until data is ready to be read
    mirf_CSN_lo; // Pull down chip select
    SPI.transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
    spi_read_data(data_in_out, mirf_PAYLOAD); // Read payload
    mirf_CSN_hi; // Pull up chip select
    mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register

    mirf_flush_rx_tx(); // Flush RX and TX FIFO

    uint8_t sensorNo = data_in_out[SENSOR_NUMBER];

    for (uint8_t x = 0; x < 14; x++) {
      Serial.print(data_in_out[x]);
      Serial.print(" ");
    }
    
    // If link byte matches
    if (data_in_out[LINK_BYTE] == linkByte) {
      sensorList[sensorNo].lastseen = currentUptime;
      sensorList[sensorNo].state = 1;
  
      uint8_t arrayCounter = 2;
      for (uint8_t x = 0; x < 6; x++) {
        double voltage = ((data_in_out[arrayCounter] << 8) | data_in_out[arrayCounter+1]) * 0.00107421875 * 26.6382;
        arrayCounter += 2;
        sensorVoltage[sensorNo][(x)] = voltage;
      }
      
      // Check for cell count
      uint8_t cells = 0;
      for (uint8_t v = 0; v < 6; v++) {
        if (sensorVoltage[sensorNo][v] >= 1) {
          cells = v+1;
        }
      }

      // Calculate voltage on cells (can't test if highest cell is 0V)
      for (uint8_t v = 0; v < cells; v++) {
        if (v >= 1) {
          for (int8_t c = v-1; c >= 0; c--) {
            sensorVoltage[sensorNo][v] -= sensorVoltage[sensorNo][c];
          }
        }
      }
    }
    
    Serial.print("Uptime ");
    Serial.print(currentUptime);
    Serial.println("");
    delay(50);
  }
}

