/*
 LiPo Battery Storage Monitor - Server
 Version: 1.0
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 19/11/2017
 Last Modified: 3/12/2017
 */

#ifndef uint8_t
 typedef unsigned char uint8_t;
#endif 

#include <SPI.h>
#include "mirf.h"
#include "nRF24L01.h"

// Write data using SPI
void spi_write_data(uint8_t * dataout, uint8_t len) {
  uint8_t i;
  for (i = 0; i < len; i++) {
    SPI.transfer(dataout[i]);
  }
}

// Read data using SPI
void spi_read_data(uint8_t * datain, uint8_t len) {
  uint8_t i;
  for (i = 0; i < len; i++) {
    datain[i] = SPI.transfer(0x00);
  }
}

// Initializes pins ans interrupt to communicate with the MiRF module
// Should be called in the early initializing phase at startup.
void mirf_init(void) {
  // Define CSN and CE as Output and set them to default
  pinMode (CSN, OUTPUT);
  pinMode (CE, OUTPUT);

  mirf_CE_lo;
  mirf_CSN_hi;
}

// Sets the important registers in the MiRF module and powers the module
// in receiving mode
void mirf_config(void) {
  // Set RF channel  
  mirf_config_register(RF_CH, mirf_CH);
  
  // Set length of incoming payload 
  mirf_config_register(RX_PW_P0, mirf_PAYLOAD);
  
  // Disable auto acknowledgement
  mirf_config_register(EN_AA, 0<<ENAA_P0);
  
  // Disable auto retransmit delay
  mirf_config_register(SETUP_RETR, 0);
  
  // Set RADDR and TADDR
  mirf_write_register(RX_ADDR_P0, RADDR, 5);
  mirf_write_register(TX_ADDR, TADDR, 5);
  
  // Enable RX_ADDR_P0 address matching
  mirf_config_register(EN_RXADDR, (1<<ERX_P0));
}

// Flush RX and TX FIFO
void mirf_flush_rx_tx(void) {
  mirf_CSN_lo; // Pull down chip select
  SPI.transfer(FLUSH_RX); // Flush RX
  mirf_CSN_hi; // Pull up chip select

  mirf_CSN_lo; // Pull down chip select
  SPI.transfer(FLUSH_TX);  // Write cmd to flush tx fifo
  mirf_CSN_hi; // Pull up chip select
}

// Checks if data is available for reading
uint8_t mirf_data_ready(void) {
  mirf_CSN_lo;                                // Pull down chip select
  SPI.transfer(R_REGISTER | (REGISTER_MASK & STATUS));
  uint8_t status = SPI.transfer(0x00);        // Read status register
  mirf_CSN_hi;                                // Pull up chip select
  return status & (1<<RX_DR);
}

// Read the status register
uint8_t mirf_status(void) {
  mirf_CSN_lo; // Pull down chip select
  SPI.transfer(R_REGISTER | (REGISTER_MASK & STATUS));
  uint8_t status = SPI.transfer(NOP); // Read status register
  mirf_CSN_hi; // Pull up chip select
  return status;
}

// Write one byte into the given MiRF register
void mirf_config_register(uint8_t reg, uint8_t value) {
  mirf_CSN_lo;
  SPI.transfer(W_REGISTER | (REGISTER_MASK & reg));
  SPI.transfer(value);
  mirf_CSN_hi;
}

// Reads an array of bytes from the given start position in the MiRF registers.
void mirf_read_register(uint8_t reg, uint8_t *value, uint8_t len) {
  mirf_CSN_lo;
  SPI.transfer(R_REGISTER | (REGISTER_MASK & reg));
  spi_read_data(value, len);
  mirf_CSN_hi;
}

// Writes an array of bytes into inte the MiRF registers
void mirf_write_register(uint8_t reg, uint8_t *value, uint8_t len) {
  mirf_CSN_lo;
  SPI.transfer(W_REGISTER | (REGISTER_MASK & reg));
  spi_write_data(value, len);
  mirf_CSN_hi;
}

uint8_t mirf_read_reg(uint8_t reg) {
  mirf_CSN_lo;
  SPI.transfer(R_REGISTER | (REGISTER_MASK & reg));
  uint8_t value = SPI.transfer(0x00);
  mirf_CSN_hi;
  return value;
}

void mirf_setup(void) {
  SPI.setFrequency(1000000);
  SPI.begin();
  mirf_init();
  delay(50);
  mirf_config();

  NRF_RX_POWERUP; // Power up to receiver mode
  delay(3);
  mirf_flush_rx_tx(); // Flush RX and TX FIFO
  mirf_CE_hi; // Start listening
}

